/* zdbf.c

   Copyright © 2011, 2012, 2013 Brandon Invergo <brandon@invergo.net>

   This file is part of zeptodb

   zeptodb is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   zeptodb is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with zeptodb.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "zdbf.h"

const char *argp_program_version = "zdbf (zeptodb 1.1)";
const char *argp_program_bug_address = "<brandon@invergo.net>";

static char doc[] =
  "zdbf -- a tool for fetching records from a DBM database\v"
  "If an input file is provided, lines are read from it as individual queries, "
  "otherwise queries are taken from stdin.";

static char args_doc[] = "DATABASE";

static struct argp_option options[] = {
  {"input", 'i', "FILE", 0, "Read queries from a file"},
  {"all", 'a', 0, 0, "Fetch all the records in the database"},
  {"delim", 'd', "CHAR", 0, "Character to print between keys and values."
   "Keys will not be printed if left unspecified."},
  {0}
};

struct arguments
{
  char *args[1];
  char *input_file;
  int all;
  char *delim;
};

static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  struct arguments *arguments = state->input;

  switch (key)
    {
    case 'i':
      arguments->input_file = arg;
      break;

    case 'a':
      arguments->all = 1;
      break;

    case 'd':
      if (strlen (arg) > 1)
        argp_usage (state);
      arguments->delim = arg;
      break;

    case ARGP_KEY_ARG:
      if (state->arg_num >= 1)
        argp_usage (state);
      arguments->args[state->arg_num] = arg;
      break;

    case ARGP_KEY_END:
      if (state->arg_num == 0)
        argp_usage (state);
      break;

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

static struct argp argp = { options, parse_opt, args_doc, doc };

void
printrec (char *key, char *value, char *delim)
{
  if (delim)
    printf ("%s%s%s\n", key, delim, value);
  else
    printf ("%s\n", value);
}

int
fetchall (GDBM_FILE database, char *delim)
{
  /*
   * This function fetches all the records from the database.
   */

  datum key, nextkey, value;
  char *key_str, *value_str;
  key = gdbm_firstkey (database);
  
  while (key.dptr)
    {
      value = gdbm_fetch (database, key);
      if (value.dptr == NULL)
        error (0, errno, "Key with non-existent value: %s", key.dptr);
      value_str = strndup (value.dptr, value.dsize);
      key_str = strndup (key.dptr, key.dsize);
      printrec (key_str, value_str, delim);
      free (value.dptr);
      free (key_str);
      free (value_str);
      nextkey = gdbm_nextkey (database, key);
      free (key.dptr);
      key = nextkey;
    }
  return (0);
}

int
fetchqueries (GDBM_FILE database, FILE * input, char *delim)
{
  /* 
   * This function fetches records from the database from queries
   * in a file object.
   */

  size_t linelen = 0;
  ssize_t chars_read;
  char *lineptr = (char *) malloc (linelen);
  char *value_str;

  // query the database for each line of input
  while (1)
    {
      datum key, value;
      chars_read = getline (&lineptr, &linelen, input);
      if (chars_read == EOF)
        break;
      if (strrchr(lineptr, '\n'))
        {
          key.dptr = strndup (lineptr, chars_read - 1);
          key.dsize = chars_read - 1;
        }
      else
        {
          key.dptr = lineptr;
          key.dsize = chars_read;
        }
      value = gdbm_fetch (database, key);
      if (value.dptr == NULL)
        error (0, errno, "Key with non-existent value: %s", key.dptr);
      value_str = strndup (value.dptr, value.dsize);
      printrec (key.dptr, value_str, delim);
      free (value.dptr);
      free (key.dptr);
    }
  return (0);
}

int
main (int argc, char **argv)
{
  GDBM_FILE database;
  FILE *input;
  struct arguments arguments;
  arguments.input_file = "-";
  arguments.all = 0;
  arguments.delim = NULL;

  argp_parse (&argp, argc, argv, 0, 0, &arguments);

  // open the database
  database = gdbm_open (arguments.args[0], 512, GDBM_READER,
                        S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP, NULL);
  if (!database)
    error (EXIT_FAILURE, errno,
           "Failed to open database %s", arguments.args[0]);

  // if the "all" argument was passed, output all the records
  // and exit
  if (arguments.all != 0)
    {
      fetchall (database, arguments.delim);
      gdbm_close (database);
      exit (EXIT_SUCCESS);
    }

  // if an input file was provided, use it, otherwise use stdin
  if (strcmp (arguments.input_file, "-") == 0)
    input = stdin;
  else
    input = fopen (arguments.input_file, "r");
  if (!input)
    error (EXIT_FAILURE, errno,
           "Failed to open input file %s", arguments.input_file);
  fetchqueries (database, input, arguments.delim);

  fclose (input);
  gdbm_close (database);
  exit (EXIT_SUCCESS);
}
