/* zdbr.c

   Copyright © 2011, 2012, 2013 Brandon Invergo <brandon@invergo.net>
   
   This file is part of zeptodb
   
   zeptodb is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.
   
   zeptodb is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with zeptodb.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "zdbr.h"

const char *argp_program_version = "zdbr (zeptodb 1.1)";
const char *argp_program_bug_address = "<brandon@invergo.net>";

static char doc[] =
  "zdbr -- a tool for removing records from a DBM database\v"
  "If an input file is provided, lines are read from it as individual queries, "
  "otherwise queries are taken from stdin. Note that if you have removed "
  "many records from the database, some fragmentation can occur, in which case "
  "it is recommended to reorganize the database. If the 'reorganize' argument "
  "is passed, the database will be reorganized after any records have been "
  "removed";

static char args_doc[] = "DATABASE";

static struct argp_option options[] = {
  {"input", 'i', "FILE", 0, "Read queries from a file"},
  {"reorganize", 'r', 0, 0, "Reorganize the database"},
  {0}
};

struct arguments
{
  char *args[1];
  char *input_file;
  int reorg;
};

static error_t
parse_opt (int key, char *arg, struct argp_state *state)
{
  struct arguments *arguments = state->input;

  switch (key)
    {
    case 'i':
      arguments->input_file = arg;
      break;

    case 'r':
      arguments->reorg = 1;
      break;

    case ARGP_KEY_ARG:
      if (state->arg_num >= 1)
        argp_usage (state);
      arguments->args[state->arg_num] = arg;
      break;

    case ARGP_KEY_END:
      if (state->arg_num == 0)
        argp_usage (state);
      break;

    default:
      return ARGP_ERR_UNKNOWN;
    }
  return 0;
}

static struct argp argp = { options, parse_opt, args_doc, doc };

int
remrecords (GDBM_FILE database, FILE * input)
{
  /* 
   * This function removes records from the database from queries
   * in a file object.
   */

  datum key;
  size_t linelen = 512;
  size_t keylen;
  char *lineptr = (char *) malloc (linelen);
  char *key_str;
  int res;

  // query the database for each line of input
  while (getline (&lineptr, &linelen, input) != EOF)
    {
      keylen = strlen (lineptr);
      if (lineptr[keylen - 1] == '\n')
        {
          key_str = (char *) malloc (sizeof (char) * (keylen));
          if (!key_str)
            error (EXIT_FAILURE, errno, "Virtual memory exhausted");
          // get rid of the newline
          strncpy (key_str, lineptr, sizeof (char) * (keylen - 1));
          strncpy (key_str + keylen - 1, "\0", sizeof (char));
          key.dptr = key_str;
          key.dsize = keylen - 1;
        }
      else
        {
          key.dptr = lineptr;
          key.dsize = keylen;
        }
      res = gdbm_delete (database, key);
      if (res < 0)
        printf ("Key not found: %s\n", key.dptr);
      free (key.dptr);
    }
  return (0);
}

int
main (int argc, char **argv)
{
  GDBM_FILE database;
  FILE *input;
  int res;
  struct arguments arguments;
  arguments.input_file = "-";
  arguments.reorg = 0;

  argp_parse (&argp, argc, argv, 0, 0, &arguments);

  // open the database
  database = gdbm_open (arguments.args[0], 512, GDBM_WRITER,
                        S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP, NULL);
  if (!database)
    error (EXIT_FAILURE, errno,
           "Failed to open database %s", arguments.args[0]);

  // if an input file was provided, use it, otherwise use stdin
  if (strcmp (arguments.input_file, "-") == 0)
    input = stdin;
  else
    input = fopen (arguments.input_file, "r");
  if (!input)
    error (EXIT_FAILURE, errno,
           "Failed to open file %s", arguments.input_file);
  remrecords (database, input);

  // if the "reorganize" argument was passed, reorganize 
  // the database
  if (arguments.reorg != 0)
    {
      res = gdbm_reorganize (database);
      if (res < 0)
        error (EXIT_FAILURE, errno, "GDBM error");
    }

  fclose (input);
  gdbm_close (database);
  exit (EXIT_SUCCESS);
}
